#ifndef _TAGS_H_

#define _TAGS_H_
#include <linux/version.h>
#include <linux/hardirq.h> /* for in_interrupt() */
#include "lve_debug_events.h"

#define SWITCH_MAGIC	0x53575645

struct switch_data {
	uint32_t		sw_magic;
	struct module		*sw_owner;
	uint32_t		sw_cookie; /* cookie used for enter */
	struct light_ve		*sw_from;
	uint32_t		sw_flags;
	uint32_t		sw_failmask;
	struct task_struct	*sw_task;
	atomic_t		sw_refcnt;
	struct lve_namespace	sw_ns;
	struct list_head	sw_list; /* for delayed release */
};

struct task_struct;
struct switch_data * switch_tag_attach(struct task_struct *task,
				       struct light_ve *ve, uint32_t cookie);
void switch_free(struct switch_data *sw_data);
void switch_tag_fork(struct task_struct * task);
void switch_tag_release(const struct switch_data *sw_data);

#if RHEL_MAJOR == 0
extern struct switch_data *lve_pid_to_swdata[PID_MAX_LIMIT];
#define LVE_TAG(tsk)	(lve_pid_to_swdata[tsk->pid])
#elif RHEL_MAJOR == 9
#define LVE_TAG(tsk)	(*(struct switch_data **)&((tsk)->rh_kabi_dummy))
#elif RHEL_MAJOR == 8
/* only used by kernel with stackprotector && ppc64 */
#define LVE_TAG(tsk)	(*(struct switch_data **)&((tsk)->rh_reserved2))
#else
/* CL7 kmodlve should not be built from this branch from now on */
#error tags storage is not defined for this kernel
#endif

extern rwlock_t lve_swdata_lock;

static inline struct switch_data *LVE_TAG_GET(struct task_struct *tsk) 
{
	struct switch_data *ret;
	unsigned long flags;
	int refcnt = 0;

	read_lock_irqsave(&lve_swdata_lock, flags);

	ret = (void *)LVE_TAG(tsk);

	if (ret == NULL || ret->sw_magic != SWITCH_MAGIC) {
		ret = NULL;
		goto out;
	}

	refcnt = atomic_inc_return(&ret->sw_refcnt);
out:
	read_unlock_irqrestore(&lve_swdata_lock, flags);

	trace_lve_tag_get(tsk, ret, refcnt);

	return ret;
}

static inline void LVE_TAG_SET(struct task_struct *tsk, struct switch_data *data)
{
	unsigned long flags;

	write_lock_irqsave(&lve_swdata_lock, flags);
	LVE_TAG(tsk) = data;
	write_unlock_irqrestore(&lve_swdata_lock, flags);
}

static inline void LVE_TAG_CLEAR(struct task_struct *tsk)
{
	LVE_TAG_SET(tsk, NULL);
}

static inline void LVE_TAG_PUT(struct switch_data *sw_data)
{
	int refcnt;

	refcnt = atomic_dec_return(&sw_data->sw_refcnt);
	if (!refcnt)
		switch_free(sw_data);

	trace_lve_tag_put(sw_data, refcnt);
}
void tag_free_delayed(struct switch_data *sw_data);

static inline void LVE_TAG_PUT_DELAYED(struct switch_data *sw_data)
{
	if (atomic_dec_and_test(&sw_data->sw_refcnt))
		tag_free_delayed(sw_data);
}

int switch_init(void);
void switch_fini(void);

#endif
